#include <gsl/gsl_errno.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_odeiv2.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "headerSchools.h"


/* Reads global variables, the number of individuals, the sigmas for the mcmc parameters */
void ReadFiles1(char *address_glo, Indiv *indiv, Negs *negs, Global *global, Param *param){
    
    FILE *globalFile;
    globalFile = fopen(address_glo, "r");
    if(globalFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_glo);
        exit(2);
    }
    while(!feof(globalFile)){fscanf(globalFile, "%d %d %d %d %d %d %lf %lf %lf %lf %d %lf %lf %lf %lf %lf", 
                                    &indiv->Nindiv,
                                    &negs->Nindiv,
                                    &global->Nclusters,
                                    &global->NDAYS,
                                    &global->DAY0,
                                    &global->DAYL,
                                    &global->XI1, 
                                    &global->XI2,
                                    &global->W,
                                    &global->RHO,
                                    &global->MAX_GENTIME,
                                    &param->sigma_alpha,
                                    &param->sigma_beta,
                                    &param->sigma_gamma,
                                    &param->sigma_Gta,
                                    &param->sigma_Gtb);
    }
    fclose(globalFile);

    global->XI3 = 0.; // not really in use. 

    param->sigma_alpha = 0.001;
    param->sigma_beta = 0.01;
    param->sigma_gamma = 0.01;
    param->sigma_Gta = 0.5;
    param->sigma_Gtb = 1.5;
    
    // initialization of the parameters
    param->alpha = 0.; param->beta = 0.; param->gamma = 0.; param->Gta = 0.; param->Gtb = 0.;

}

/* Reads individuals data, community incidence, and infection times */
void ReadFiles2(char *address_ind, char *address_neg, char *address_idxFam, char *address_idxSch, char *address_idxCom,
                char *address_idxComNeg, char *address_out, char *address_inf, Indiv *indiv, Negs *negs, Global global, Out *out, int ***infecTimes, int NinfSets){

    int i, j;

    // Individuals data
    i = 0;
    FILE *indivFile;
    indivFile = fopen(address_ind, "r");
    if(indivFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_ind);
        exit(2);
    }
    while(!feof(indivFile)){
        fscanf(indivFile,"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d", 
               &indiv->IDperson[i],
               &indiv->IDfamily[i],
               &indiv->IDcluster[i],
               &indiv->symonset[i],
               &indiv->diagnose[i], 
               &indiv->neg1[i],
               &indiv->neg2[i],
               &indiv->neg3[i],
               &indiv->neg4[i],
               &indiv->neg5[i],
               &indiv->neg6[i],
               &indiv->neg7[i],
               &indiv->neg8[i],
               &indiv->neg9[i],
               &indiv->vax1[i],
               &indiv->vax2[i],
               &indiv->vax3[i],
               &indiv->attend[i]); 
        i++;
    }
    fclose(indivFile);


    // Negative individuals data
    i = 0;
    FILE *negFile;
    negFile = fopen(address_neg, "r");
    if(negFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_neg);
        exit(2);
    }
    while(!feof(negFile)){
        fscanf(negFile,"%d %d %d %d %d", 
               &negs->IDperson[i],
               &negs->IDfamily[i],
               &negs->vax1[i],
               &negs->vax2[i],
               &negs->vax3[i]); 
        i++;
    }
    fclose(negFile);


    // Index functions
    i = 0; j = 0;
    FILE *indexFamFile;
    indexFamFile = fopen(address_idxFam, "r");
    if(indexFamFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_idxFam);
        exit(2);
    }
    while(!feof(indexFamFile)){
        if(j < global.NDAYS){
            fscanf(indexFamFile, "%d", &((indiv->idxFam)[i][j]));
            j++;
        }
        else{i++; j=0;}
    }
    fclose(indexFamFile);

    i = 0; j = 0;
    FILE *indexSchFile;
    indexSchFile = fopen(address_idxSch, "r");
    if(indexSchFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_idxSch);
        exit(2);
    }
    while(!feof(indexSchFile)){
        if(j < global.NDAYS){
            fscanf(indexSchFile, "%d", &((indiv->idxSch)[i][j]));
            j++;
        }
        else{i++; j=0;}
    }
    fclose(indexSchFile);

    i = 0; j = 0;
    FILE *indexComFile;
    indexComFile = fopen(address_idxCom, "r");
    if(indexComFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_idxCom);
        exit(2);
    }
    while(!feof(indexComFile)){
        if(j < global.NDAYS){
            fscanf(indexComFile, "%d", &((indiv->idxCom)[i][j]));
            j++;
        }
        else{i++; j=0;}
    }
    fclose(indexComFile);

    i = 0; j = 0;
    FILE *indexComNegFile;
    indexComNegFile = fopen(address_idxComNeg, "r");
    if(indexComNegFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_idxComNeg);
        exit(2);
    }
    while(!feof(indexComNegFile)){
        if(j < global.NDAYS){
            fscanf(indexComNegFile, "%d", &((negs->idxCom)[i][j]));
            j++;
        }
        else{i++; j=0;}
    }
    fclose(indexComNegFile);


    // Incidence in the community
    i = 0;
    FILE *outFile;
    outFile = fopen(address_out, "r");
    if(outFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_out);
        exit(2);
    }
    
    while(!feof(outFile)){
        fscanf(outFile,"%d %lf", &out->date[i], &out->incidence[i]);
        i++;
    }
    fclose(outFile);

    double max_incidence = 0.;
    for(i=0; i<global.NDAYS; i++){
        if(out->incidence[i] > max_incidence){
            max_incidence = out->incidence[i];
        }
    }
    for(i=0; i<global.NDAYS; i++){
        out->incidence[i] = out->incidence[i] / max_incidence;
    }


    // Sets of infection times
    i = 0; j = 0;
    FILE *infecFile;
    infecFile = fopen(address_inf,"r");
    if(infecFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_inf);
        exit(2);
    }
    while(!feof(infecFile) && (i < NinfSets)){
        if(j < indiv->Nindiv){
            fscanf(infecFile, "%d", &((*infecTimes)[i][j]));
            j++;
        } 
        else{i++; j=0;}
    }
    fclose(infecFile);

}

/* Assigns the daily susceptibility to all positive individuals, from day 0 to last day (412) */
void Xi(Indiv *indiv, Global global){

    int i, j, lim1, lim2, lim3;

    for(i=0; i<indiv->Nindiv; i++){

        if(indiv->vax1[i] != -1){lim1 = indiv->vax1[i]+14;} else{lim1 = -1;}
        if(indiv->vax2[i] != -1){lim2 = indiv->vax2[i]+14;} else{lim2 = -1;}
        if(indiv->vax3[i] != -1){lim3 = indiv->vax3[i]+14;} else{lim3 = -1;}

        for(j=0; j<global.NDAYS; j++){
            
            indiv->xi[i][j] = 1.;

            if(lim1 != -1){
                if(j >= lim1){
                    indiv->xi[i][j] = 1 - global.XI1 * exp(-global.W * (j - indiv->vax1[i] - 14));
                }
                if(lim2 != -1){
                    if(j >= lim2){
                        indiv->xi[i][j] = 1 - global.XI2 * exp(-global.W * (j - indiv->vax2[i] - 14));
                    }
                    if(lim3 != -1){
                        if(j >= lim3){
                            indiv->xi[i][j] = 1 - global.XI3;
                        }
                    }
                }
            }

        }
    }

}

/* Assigns the daily susceptibility to all negative individuals, from day 0 to last day (412) */
void XiNeg(Negs *negs, Global global){

    int i, j, lim1, lim2, lim3;

    for(i=0; i<negs->Nindiv; i++){

        if(negs->vax1[i] != -1){lim1 = negs->vax1[i]+14;} else{lim1 = -1;}
        if(negs->vax2[i] != -1){lim2 = negs->vax2[i]+14;} else{lim2 = -1;}
        if(negs->vax3[i] != -1){lim3 = negs->vax3[i]+14;} else{lim3 = -1;}

        for(j=0; j<global.NDAYS; j++){
            
            negs->xi[i][j] = 1.;

            if(lim1 != -1){
                if(j >= lim1){
                    negs->xi[i][j] = 1 - global.XI1 * exp(-global.W * (j - negs->vax1[i] - 14));
                }
                if(lim2 != -1){
                    if(j >= lim2){
                        negs->xi[i][j] = 1 - global.XI2 * exp(-global.W * (j - negs->vax2[i] - 14));
                    }
                    if(lim3 != -1){
                        if(j >= lim3){
                            negs->xi[i][j] = 1 - global.XI3;
                        }
                    }
                }
            }

        }
    }

}

/* Assigns the daily transmissibility to all positive individuals, from day 0 to last day (412) */
void Rho(Indiv *indiv, Global global){

    int i, j, lim1;

    for(i=0; i<indiv->Nindiv; i++){

        if(indiv->vax1[i] != -1){lim1 = indiv->vax1[i]+14;} else{lim1 = -1;}

        for(j=0; j<global.NDAYS; j++){
            
            indiv->rho[i][j] = 1.;

            if(lim1 != -1){
                if(j >= lim1){indiv->rho[i][j] = global.RHO;}
            }

        }
    }

}

/* Computes the vector for the generation time density evaluated from 0 to MAX_GENTIME given the parameters */
void GenDensity(Param *param, Global global){
    
    int i, j;

    for(i=0; i<=global.MAX_GENTIME; i++){
        param->genDensity[i] = 0.;
        for(j=0; j<steps; j++){
            param->genDensity[i] += gsl_ran_gamma_pdf((double)(i+j/steps), ((param->Gta)*(param->Gta)/(param->Gtb)), ((param->Gtb)/(param->Gta)))/steps;
        }
    }
    
}

/* Computes the different lambda's (forces of infection) from the community and from each household/school individual (for positive individuals) */
long double **ForcesInfec_OHS(Indiv indiv, Global global, Out out, Param param, int **infecTimes, int person, int setOfInfecTimes){
                                                                                  
    int i, z, timeInfec;
    long double p_outside, *prob_houseMemb, *prob_schoolMemb, **probs;
    prob_houseMemb = (long double *)calloc(indiv.Nindiv, sizeof(long double));
    prob_schoolMemb = (long double *)calloc(indiv.Nindiv, sizeof(long double));
    timeInfec = infecTimes[setOfInfecTimes][person];

    probs = (long double **)calloc(3, sizeof(long double *));
    probs[0] = (long double *)calloc(1, sizeof(long double));
    probs[1] = (long double *)calloc(indiv.Nindiv, sizeof(long double));
    probs[2] = (long double *)calloc(indiv.Nindiv, sizeof(long double));
    

    // Force of infection from community (place = 0)
    p_outside = 0.;
    for(z=global.MINT; z<=timeInfec; z++){
        if(timeInfec - z <= global.MAX_GENTIME){
            p_outside += out.incidence[z] * indiv.xi[person][timeInfec] * param.genDensity[(int)(timeInfec-z)] * indiv.idxCom[person][timeInfec];
        }
    }
    p_outside = param.alpha * p_outside;
    probs[0][0] = p_outside;

    // Force of infection from the household (place = 1)
    if(indiv.IDfamily[person] == -1){} else{
        for(i=0; i<indiv.Nindiv; i++){
            prob_houseMemb[i] = 0.;
            if(person != i && indiv.IDfamily[i] == indiv.IDfamily[person] && infecTimes[setOfInfecTimes][i] <= timeInfec && (timeInfec-infecTimes[setOfInfecTimes][i]) <= global.MAX_GENTIME){ 
                prob_houseMemb[i] = param.beta * indiv.rho[i][timeInfec] * indiv.xi[person][timeInfec] * param.genDensity[(int)(timeInfec-infecTimes[setOfInfecTimes][i])];
            }
            probs[1][i] = prob_houseMemb[i];
        }
    }

    // Force of infection from the school (place = 2)
    if(indiv.attend[person] == 0){} else{
        for(i=0; i<indiv.Nindiv; i++){
            prob_schoolMemb[i] = 0.;
            if(person != i && indiv.IDcluster[i] == indiv.IDcluster[person] && indiv.attend[i] == 1 && infecTimes[setOfInfecTimes][i] <= timeInfec && (timeInfec-infecTimes[setOfInfecTimes][i]) <= global.MAX_GENTIME){
                prob_schoolMemb[i] = param.gamma * indiv.rho[i][timeInfec] * indiv.xi[person][timeInfec] * param.genDensity[(int)(timeInfec-infecTimes[setOfInfecTimes][i])] * indiv.idxSch[person][timeInfec] * indiv.idxSch[i][timeInfec];
            }
            probs[2][i] = prob_schoolMemb[i];
        }
    }

    free(prob_houseMemb);
    free(prob_schoolMemb);

    return(probs);

}

/* Computes the integral (positive) of the exponential in the Q term of the likelihood, for the positive individuals */
long double Qp_OHS(Indiv indiv, Global global, Out out, Param param, int **infecTimes, int person, int setOfInfecTimes){
                                                                                  
    int i, z, t, timeInfec;
    long double cum_outside, cum_house, cum_school;
    
    cum_outside = 0.; cum_house = 0.; cum_school = 0.;
    timeInfec = infecTimes[setOfInfecTimes][person];
    
    // Force of infection from community - cummulative
    for(t=global.MINT; t<=timeInfec; t++){
        for(z=global.MINT; z<=t; z++){
            if(t - z <= global.MAX_GENTIME){
                cum_outside += out.incidence[z] * indiv.xi[person][t] * param.genDensity[(int)(t-z)] * indiv.idxCom[person][t];
            }
        }
    }
    cum_outside = param.alpha * cum_outside;

    // Force of infection from the household - cumulative
    if(indiv.IDfamily[person] == -1){} else{
        for(i=0; i<indiv.Nindiv; i++){
            if(person != i && indiv.IDfamily[i] == indiv.IDfamily[person] && infecTimes[setOfInfecTimes][i] <= timeInfec){ 
                for(t=infecTimes[setOfInfecTimes][i]; t<=timeInfec; t++){
                    if(t-infecTimes[setOfInfecTimes][i] <= global.MAX_GENTIME){
                        cum_house += indiv.rho[i][t] * indiv.xi[person][t] * param.genDensity[(int)(t-infecTimes[setOfInfecTimes][i])];
                    }
                }
            }
        }
    }
    cum_house = param.beta * cum_house;

    // Force of infection from the school - cumulative
    if(indiv.attend[person] == 0){} else{
        for(i=0; i<indiv.Nindiv; i++){
            if(person != i && indiv.IDcluster[i] == indiv.IDcluster[person] && indiv.attend[i] == 1 && infecTimes[setOfInfecTimes][i] <= timeInfec){
                for(t=infecTimes[setOfInfecTimes][i]; t<=timeInfec; t++){
                    if(t-infecTimes[setOfInfecTimes][i] <= global.MAX_GENTIME){
                        cum_school += indiv.rho[i][t] * indiv.xi[person][t] * param.genDensity[(int)(t-infecTimes[setOfInfecTimes][i])] * indiv.idxSch[person][t] * indiv.idxSch[i][t];
                    }
                }
            }
        }
    }
    cum_school = param.gamma * cum_school;

    return(cum_outside + cum_house + cum_school);
}

/* Computes the integral (positive) of the exponential in the Q term of the likelihood, for the negative family individuals */
long double Qn_OH(Indiv indiv, Negs negs, Global global, Out out, Param param, int **infecTimes, int person, int setOfInfecTimes){
                                                                                  
    int i, z, t;
    long double cum_outside, cum_house;
    
    cum_outside = 0.; cum_house = 0.;
    
    // Force of infection from the community - cummulative
    for(t=global.MINT; t<=global.MAXT; t++){
        for(z=global.MINT; z<=t; z++){
            if(t - z <= global.MAX_GENTIME){
                cum_outside += out.incidence[z] * negs.xi[person][t] * param.genDensity[(int)(t-z)] * negs.idxCom[person][t];
            }
        }
    }
    cum_outside = param.alpha * cum_outside;

    // Force of infection from the household - cumulative
    for(i=0; i<indiv.Nindiv; i++){
        if(indiv.IDfamily[i] == negs.IDfamily[person]){ 
            for(t=infecTimes[setOfInfecTimes][i]; t<=global.MAXT; t++){
                if(t-infecTimes[setOfInfecTimes][i] <= global.MAX_GENTIME){
                    cum_house += indiv.rho[i][t] * negs.xi[person][t] * param.genDensity[(int)(t-infecTimes[setOfInfecTimes][i])];
                }
            }
        }
    }
    cum_house = param.beta * cum_house;

    return(cum_outside + cum_house);
}

/* Computes the log-likelihood given the parameter set */
long double Likelihood(Indiv indiv, Negs negs, Global global, Out out, Param param, Chain *chain, int **infecTimes, int setOfInfecTimes){

    long double **probs, *Qp, Qn;
    long double loglike, p_outside, p_house, p_school, pbest_house, pbest_school;
    int i, j, ok, ibest_house, ibest_school;

    Qp = (long double *)calloc(indiv.Nindiv, sizeof(long double));
    
    loglike = 0.;

    // Contribution to the log-likelihood of the negative family members
    for(j=0; j<negs.Nindiv; j++){
        Qn = Qn_OH(indiv, negs, global, out, param, infecTimes, j, setOfInfecTimes);
        loglike += log10l((long double)exp(-Qn));
    }
     
    // Contribution to the log-likelihood of the positive family members
    for(j=0; j<indiv.Nindiv; j++){

        if(indiv.IDfamily[j] == -1){ // Does not contribute to the likelihood if he has no family.
            Qp[j] = 0.;
            chain->probInf[j] = 1.;
            chain->infector[j] = -999;
            chain->placeInf[j] = -999;

        } else{
            Qp[j] = Qp_OHS(indiv, global, out, param, infecTimes, j, setOfInfecTimes);
            probs = ForcesInfec_OHS(indiv, global, out, param, infecTimes, j, setOfInfecTimes);

            p_outside = probs[0][0];
            if(p_outside <= 1e-299){p_outside = 1e-299;}

            pbest_house = 1e-299; ibest_house = -1;
            pbest_school = 1e-299; ibest_school = -1;

            for(i=0; i<indiv.Nindiv; i++){
                p_house = probs[1][i];
                p_school = probs[2][i];
                if(p_house > pbest_house){
                    pbest_house = p_house;
                    ibest_house = i;
                }
                if(p_school > pbest_school){
                    pbest_school = p_school;
                    ibest_school = i;
                }
            }

            free(probs[0]); 
            free(probs[1]); 
            free(probs[2]); 
            free(probs);
            
            ok = 0;

            if(p_outside >= pbest_house && p_outside >= pbest_school){
                chain->probInf[j] = p_outside; 
                chain->infector[j] = -1; 
                chain->placeInf[j] = 0;
                ok += 1;
            }
            else if(pbest_house >= pbest_school && pbest_house >= p_outside && ibest_house != -1){
                chain->probInf[j] = pbest_house; 
                chain->infector[j] = ibest_house; 
                chain->placeInf[j] = 1;
                ok += 1;
            }
            else if(pbest_school >= pbest_house && pbest_school >= p_outside && ibest_school != -1){
                chain->probInf[j] = pbest_school; 
                chain->infector[j] = ibest_school; 
                chain->placeInf[j] = 2;
                ok += 1;
            }

            if(ok == 1){} else{printf("\nNot able to assign infector\n"); exit(3);}
        

            // Check that infector-infectee do not infect each other (assign new one to outdoor infection)
            int s; s = chain->infector[j];
            if(s != -1 && s != -999 && s < j && chain->infector[s] == j){
                chain->probInf[j] = p_outside;
                chain->infector[j] = -1;
                chain->placeInf[j] = 0;
            }
        }
    }


    for(j=0; j<indiv.Nindiv; j++){
        loglike += log10l(chain->probInf[j]) + log10l((long double)exp(-Qp[j])); 
    }

    free(Qp);

    return(loglike);
}

/* Assigns an infector to each individual (with family) */
void ChainRec(Indiv indiv, Global global, Out out, Param param, Chain *chain, int **infecTimes, int setOfInfecTimes){

    int i, j, s, ok, select;
    double coin, *weightVector;
    long double **probs;
    gsl_ran_discrete_t *probVector;

    for(j=0; j<indiv.Nindiv; j++){

        if(indiv.IDfamily[j] == -1){
            //chain->probInf[j] = -999;
            chain->infector[j] = -999;
            chain->placeInf[j] = -999;

        } else{
            weightVector = (double *)calloc(2*indiv.Nindiv+1, sizeof(double));
            probs = ForcesInfec_OHS(indiv, global, out, param, infecTimes, j, setOfInfecTimes);

            for(i=0; i<indiv.Nindiv; i++){weightVector[i] = probs[1][i];} // house
            for(i=0; i<indiv.Nindiv; i++){weightVector[indiv.Nindiv+i] = probs[2][i];} // school
            weightVector[2*indiv.Nindiv] = probs[0][0]; // community

            probVector = (gsl_ran_discrete_t *)calloc(2*indiv.Nindiv+1, sizeof(gsl_ran_discrete_t));
            probVector = gsl_ran_discrete_preproc(2*indiv.Nindiv+1, weightVector);
            select = gsl_ran_discrete(RNG, probVector);
            gsl_ran_discrete_free(probVector);

            ok = 0;
            for(i=0; i<(2*indiv.Nindiv+1); i++){
                if(weightVector[i] > 1e-100){ok += 1;} // To avoid sampling from a vector of probability 0
            }

            if(ok == 0){
                chain->infector[j] = -1;
                chain->placeInf[j] = 0;
            } 
            else{

                ok = 0;

                if(select == 2*indiv.Nindiv){ // community infection
                    //chain->probInf[j] = weightVector[select];
                    chain->infector[j] = -1;
                    chain->placeInf[j] = 0; 
                    ok += 1;
                }
                else if(select < indiv.Nindiv){ // household infection
                    //chain->probInf[j] = weightVector[select];
                    chain->infector[j] = select; 
                    chain->placeInf[j] = 1; 
                    ok += 1;
                }
                else if(select >= indiv.Nindiv && select < 2*indiv.Nindiv){ // school infection
                    //chain->probInf[j] = weightVector[select];
                    chain->infector[j] = select - indiv.Nindiv; 
                    chain->placeInf[j] = 2; 
                    ok += 1;
                }

                if(ok == 1){} else{printf("\nNot able to assign infector when reconstructing the chain\n"); exit(3);}

                // Avoiding infector-infectee infected each other
                s = chain->infector[j];
                if(s != -1 && s != -999 && s < j && chain->infector[s] == j){
                    coin = gsl_ran_flat(RNG,0,1);
                    if(coin < 0.5){
                        chain->infector[j] = -1;
                        chain->placeInf[j] = 0;
                    } else{
                        chain->infector[s] = -1;
                        chain->placeInf[s] = 0;
                    }
                }
            }

            free(probs[0]); 
            free(probs[1]); 
            free(probs[2]); 
            free(probs);
            free(weightVector);

        }
    }
    
}
