#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_odeiv2.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <math.h>
#include <time.h>
#include "headerSchools.h"


int main(int argc, char *argv[]){


    // Command line and basics ---------------------------------------------------------------------------------------------------//

    int Nsim, burnin, NinfSets;
    char *Nsim_, *burnin_, *NinfSets_, *seed; 
    char *address_glo, *address_ind, *address_neg, *address_idxFam, *address_idxSch, *address_idxCom, *address_idxComNeg, *address_out, *address_inf;
    char help[] = "\nERROR COMMAND LINE\nusage: ./execute/executable -sm [Nsim] -bn [BurnIn] -sd [seed] -ag [address_glo] -ai [address_ind] -an [address_neg] -axf [address_idxFam] -axs [address_idxSch] -axc [address_indexCom] -axcn [address_indexComNeg] -ao [address_outIncidence] -ainf [address_infectiontimes] -ninft [Ninft]\n";
    
    if(argc < 27){fprintf(stderr,"%s\n",help); exit(1);}

    Nsim_ = argv[2]; if(Nsim_ == NULL){Nsim_ = "0";}
    burnin_ = argv[4]; if(burnin_ == NULL){burnin_ = "0";}
    Nsim = atoi(Nsim_);
    burnin = atoi(burnin_);
    seed = argv[6]; if(seed == NULL){seed = "0";}
    address_glo = argv[8]; if(address_glo == NULL){address_glo = "0";}
    address_ind = argv[10]; if(address_ind == NULL){address_ind = "0";}
    address_neg = argv[12]; if(address_neg == NULL){address_neg = "0";}
    address_idxFam = argv[14]; if(address_idxFam == NULL){address_idxFam = "0";}
    address_idxSch = argv[16]; if(address_idxSch == NULL){address_idxSch = "0";}
    address_idxCom = argv[18]; if(address_idxCom == NULL){address_idxCom = "0";}
    address_idxComNeg = argv[20]; if(address_idxComNeg == NULL){address_idxComNeg = "0";}
    address_out = argv[22]; if(address_out == NULL){address_out = "0";}
    address_inf = argv[24]; if(address_inf == NULL){address_inf = "0";}
    NinfSets_ = argv[26]; if(NinfSets_ == NULL){NinfSets_ = "0";}
    NinfSets = atoi(NinfSets_);
    
    // Output address:
    char variant[10] = {}; 
    memcpy(variant,&address_ind[18],5);
    char address_output[100] = {}; 
    strcat(address_output, "output/mcmc.");
    strcat(address_output, variant);
    strcat(address_output, ".txt");

    // Seed:
    setenv("GSL_RNG_SEED", seed, 1);
    gsl_rng_env_setup();
    RNG = gsl_rng_alloc(gsl_rng_default);


    // Variables: ----------------------------------------------------------------------------------------------------------------//
   
    int i;
    Indiv indiv;
    Negs negs;
    Out out;
    Global global;
    Param param;
    Chain chain;
    int **infecTimes;

    int it, batch, accepted, aux1, aux2;
    double cumulativeGt, u;
    long double currentLogLike, candidateLogLike, quotient;
    long double current_alpha, current_beta, current_gamma, current_Gta, current_Gtb;


    // Allocation and Reading Files: ---------------------------------------------------------------------------------------------//

    ReadFiles1(address_glo, &indiv, &negs, &global, &param);

    param.genDensity = (double *)calloc(global.MAX_GENTIME+1, sizeof(double));

    indiv.IDperson = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.IDfamily = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.IDcluster = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.symonset = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.diagnose = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.neg1 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.neg2 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.neg3 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.neg4 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.neg5 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.neg6 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.neg7 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.neg8 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.neg9 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.vax1 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.vax2 = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.vax3 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.attend = (int *)calloc(indiv.Nindiv, sizeof(int));
    
    indiv.idxCom = (int **)calloc(indiv.Nindiv, sizeof(int *));
    for(i=0; i<indiv.Nindiv; i++){
        (indiv.idxCom)[i] = (int *)calloc(global.NDAYS, sizeof(int));
    }
    indiv.idxFam = (int **)calloc(indiv.Nindiv, sizeof(int *));
    for(i=0; i<indiv.Nindiv; i++){
        (indiv.idxFam)[i] = (int *)calloc(global.NDAYS, sizeof(int));
    }
    indiv.idxSch = (int **)calloc(indiv.Nindiv, sizeof(int *));
    for(i=0; i<indiv.Nindiv; i++){
        (indiv.idxSch)[i] = (int *)calloc(global.NDAYS, sizeof(int));
    }
    indiv.xi = (long double **)calloc(indiv.Nindiv, sizeof(long double *));
    for(i=0; i<indiv.Nindiv; i++){
        (indiv.xi)[i] = (long double *)calloc(global.NDAYS, sizeof(long double));
    }
    indiv.rho = (double **)calloc(indiv.Nindiv, sizeof(double *));
    for(i=0; i<indiv.Nindiv; i++){
        (indiv.rho)[i] = (double *)calloc(global.NDAYS, sizeof(double));
    }

    negs.IDperson = (int *)calloc(negs.Nindiv, sizeof(int)); 
    negs.IDfamily = (int *)calloc(negs.Nindiv, sizeof(int));
    negs.vax1 = (int *)calloc(negs.Nindiv, sizeof(int)); 
    negs.vax2 = (int *)calloc(negs.Nindiv, sizeof(int)); 
    negs.vax3 = (int *)calloc(negs.Nindiv, sizeof(int));

    negs.idxCom = (int **)calloc(negs.Nindiv, sizeof(int *));
    for(i=0; i<negs.Nindiv; i++){
        (negs.idxCom)[i] = (int *)calloc(global.NDAYS, sizeof(int));
    }
    negs.xi = (long double **)calloc(negs.Nindiv, sizeof(long double *));
    for(i=0; i<negs.Nindiv; i++){
        (negs.xi)[i] = (long double *)calloc(global.NDAYS, sizeof(long double));
    }

    out.date = (int *)calloc(global.NDAYS, sizeof(int)); 
    out.incidence = (double *)calloc(global.NDAYS, sizeof(double));

    infecTimes = (int **)calloc(NinfSets, sizeof(int *));
    for(i=0; i<NinfSets; i++){
        infecTimes[i] = (int *)calloc(indiv.Nindiv, sizeof(int));
    }

    chain.probInf = (long double *)calloc(indiv.Nindiv, sizeof(long double));
    chain.infector = (int *)calloc(indiv.Nindiv, sizeof(int));
    chain.placeInf = (int *)calloc(indiv.Nindiv, sizeof(int));

    ReadFiles2(address_ind, address_neg, address_idxFam, address_idxSch, address_idxCom, address_idxComNeg,
               address_out, address_inf, &indiv, &negs, global, &out, &infecTimes, NinfSets);

    Xi(&indiv, global);
    Rho(&indiv, global); 
    XiNeg(&negs, global);


    // Computations (MCMC): ------------------------------------------------------------------------------------------------------//

    FILE *mcmc;
    mcmc = fopen(address_output,"w");
    fprintf(mcmc, "%s %s %s %s %s %s %s %s\n", "nSim", "alpha", "beta", "gamma", "Gta", "Gtb", "logL", "nAccept");

    printf("\rIn progress 0/%d", Nsim*NinfSets); fflush(stdout);

    // Iteration over the different sets (batches) of infection times
    for(batch=0; batch<NinfSets; batch++){ 

        // Initialization
        current_alpha = ini_alpha;
        current_beta = ini_beta;
        current_gamma = ini_gamma;
        current_Gta = ini_Gta;
        current_Gtb = ini_Gtb;
        currentLogLike = -999999.9;
        accepted = 0;

        // Defining the initial and last time considered in the simulations
        global.MINT = 412;
        global.MAXT = 0;
        for(i=0; i<indiv.Nindiv; i++){
            aux1 = infecTimes[batch][i];
            aux2 = indiv.diagnose[i];
            if(aux1 < global.MINT){
                global.MINT = aux1;
            }
            if(aux2 > global.MAXT){
                global.MAXT = aux2;
            }
        }
        global.MINT = global.MINT - global.MAX_GENTIME;
        
        // MCMC iterations
        for(it=0; it<Nsim; it++){

            param.alpha = current_alpha + gsl_ran_gaussian(RNG, param.sigma_alpha);
            while(param.alpha < 0){param.alpha = current_alpha + gsl_ran_gaussian(RNG, param.sigma_alpha);}

            param.beta = current_beta + gsl_ran_gaussian(RNG, param.sigma_beta);
            while(param.beta < 0){param.beta = current_beta + gsl_ran_gaussian(RNG, param.sigma_beta);}

            param.gamma = current_gamma + gsl_ran_gaussian(RNG, param.sigma_gamma);
            while(param.gamma < 0){param.gamma = current_gamma + gsl_ran_gaussian(RNG, param.sigma_gamma);}

            param.Gta = current_Gta + gsl_ran_gaussian(RNG, param.sigma_Gta);
            while(param.Gta < 0){param.Gta = current_Gta + gsl_ran_gaussian(RNG, param.sigma_Gta);}

            param.Gtb = current_Gtb + gsl_ran_gaussian(RNG, param.sigma_Gtb);
            while(param.Gtb < 0){param.Gtb = current_Gtb + gsl_ran_gaussian(RNG, param.sigma_Gtb);}

            GenDensity(&param, global);
            cumulativeGt = 0.;
            for(i=0; i<=global.MAX_GENTIME; i++){
                cumulativeGt += param.genDensity[i];
            }

            // Restricting to cases where the generation time density is well defined
            if((param.genDensity[0] < lolim_cum) && (cumulativeGt > uplim_cum)){
                candidateLogLike = Likelihood(indiv, negs, global, out, param, &chain, infecTimes, batch);
            }else{
                candidateLogLike = -9999999.9;
            }
            
            // Selection
            quotient = candidateLogLike - currentLogLike;
            u = gsl_ran_flat(RNG,0,1); u = log10(u);
        
            if(u < quotient){
                currentLogLike = candidateLogLike;
                current_alpha = param.alpha;
                current_beta = param.beta;
                current_gamma = param.gamma;
                current_Gta = param.Gta;
                current_Gtb = param.Gtb;
                accepted++;
            }

            // Saving the results
            if(it >= burnin){
                fprintf(mcmc, "%d %.10Lf %.10Lf %.10Lf %.10Lf %.10Lf %.10Lf %d\n", it+1, 
                        current_alpha, current_beta, current_gamma, current_Gta, current_Gtb, currentLogLike, accepted);
            
                fflush(mcmc);
                fflush(mcmc);
            }



            printf("\rIn progress %d/%d", Nsim*batch + it+1, Nsim*NinfSets);
            fflush(stdout);

        }
    }

    fclose(mcmc);
    printf("\n");


    // Free: ---------------------------------------------------------------------------------------------------------------------//
    
    free(param.genDensity);

    free(indiv.IDperson);
    free(indiv.IDfamily);
    free(indiv.IDcluster);
    free(indiv.symonset);
    free(indiv.diagnose);
    free(indiv.neg1);
    free(indiv.neg2);
    free(indiv.neg3);
    free(indiv.neg4);
    free(indiv.neg5);
    free(indiv.neg6);
    free(indiv.neg7);
    free(indiv.neg8); 
    free(indiv.neg9);
    free(indiv.vax1);
    free(indiv.vax2);
    free(indiv.vax3);
    free(indiv.attend); 
    for(i=0; i<indiv.Nindiv; i++){
        free(indiv.idxCom[i]);
        free(indiv.idxFam[i]);
        free(indiv.idxSch[i]);
        free(indiv.xi[i]);
        free(indiv.rho[i]);
    }
    free(indiv.idxCom); 
    free(indiv.idxFam); 
    free(indiv.idxSch); 
    free(indiv.xi); 
    free(indiv.rho);

    free(negs.IDperson);
    free(negs.IDfamily);
    free(negs.vax1);
    free(negs.vax2);
    free(negs.vax3);
    for(i=0; i<negs.Nindiv; i++){
        free(negs.idxCom[i]);
        free(negs.xi[i]);
    }
    free(negs.idxCom);
    free(negs.xi);

    free(out.date);
    free(out.incidence);

    for(i=0; i<NinfSets; i++){
        free(infecTimes[i]);
    }
    free(infecTimes);

    free(chain.probInf);
    free(chain.infector);
    free(chain.placeInf);


    return 0;
}




