#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_odeiv2.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <math.h>
#include <time.h>


// Global variables --------------------------------------------------------------------------------------------------------------//

gsl_rng * RNG;


// Structures --------------------------------------------------------------------------------------------------------------------//

typedef struct{
    int Nindiv; /* number of individuals */
    int *IDperson; /* ID of the individual */
    int *IDfamily; /* ID of the household */
    int *IDcluster; /* ID of the outbreak */
    int *symonset; /* date of symptom onset */
    int *diagnose; /* date of diagnosis */
    int *neg1; /* negative tests' dates */
    int *neg2;
    int *neg3;
    int *neg4;
    int *neg5;
    int *neg6;
    int *neg7;
    int *neg8;
    int *neg9;
    int *vax1;
    int *vax2;
    int *vax3;
    int *attend; /* attend school or not */
} Indiv;

typedef struct{
    double DISTR_INCUBATION_SHAPE;
    double DISTR_INCUBATION_SCALE;
    double DISTR_DIAGNOSIS_SHAPE;
    double DISTR_DIAGNOSIS_SCALE;
} Distr;


// Function's calls --------------------------------------------------------------------------------------------------------------//

void ReadFiles1(char *address_glo, Indiv *indiv, Distr *distr);
void ReadFiles2(char *address_ind, Indiv *indiv);
void AssignInfecTimes(int ***infecTimes, Indiv indiv, Distr distr, int Nsim, int max_interval);


//--------------------------------------------------------------------------------------------------------------------------------//
// MAIN  -------------------------------------------------------------------------------------------------------------------------//
//--------------------------------------------------------------------------------------------------------------------------------//

int main(int argc, char *argv[]){

    // Command line and basics ---------------------------------------------------------------------------------------------------//
    
    int Nsim, max_interval;
    char *Nsim_, *max_interval_, *seed, *address_glo, *address_ind;
    char help[] = "\nERROR COMMAND LINE\nusage: ./execute/executable -sm [Nsim] -mi [max-interval-infection-diagnose+1] -sd [seed] -ag [address_glo] -ai [address_ind]\n";
    // -sm: number of simulations; -mi: maximum possible interval between infection and diagnosis + 1; -sd: seed; -ag: input global data; -ai: input individual data

    if(argc < 11){fprintf(stderr,"%s\n",help); exit(1);}

    Nsim_ = argv[2]; if(Nsim_ == NULL){Nsim_ = "0";}
    max_interval_ = argv[4]; if(max_interval_ == NULL){max_interval_ = "0";}
    Nsim = atoi(Nsim_); 
    max_interval = atoi(max_interval_);
    seed = argv[6]; if(seed == NULL){seed = "0";}
    address_glo = argv[8]; if(address_glo == NULL){address_glo = "0";}
    address_ind = argv[10]; if(address_ind == NULL){address_ind = "0";}
    
    // Output address:
    char variant[10] = {}; 
    memcpy(variant,&address_ind[18],5);
    char address_output[100] = {}; 
    strcat(address_output, "output/infection.times."); 
    strcat(address_output, variant); 
    strcat(address_output, ".txt");
    
    // Seed:
    setenv("GSL_RNG_SEED", seed, 1);
    gsl_rng_env_setup();
    RNG = gsl_rng_alloc(gsl_rng_default);


    // Variables: ----------------------------------------------------------------------------------------------------------------//
   
    int i, j;
    Indiv indiv;
    Distr distr;
    int **infecTimes;


    // Allocation and Reading Files: ----------------------------------------------------------------------------------------------//
   
    ReadFiles1(address_glo, &indiv, &distr);
    
    indiv.IDperson = (int *)calloc(indiv.Nindiv, sizeof(int)); 
    indiv.IDfamily = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.IDcluster = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.symonset = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.diagnose = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.neg1 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.neg2 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.neg3 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.neg4 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.neg5 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.neg6 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.neg7 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.neg8 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.neg9 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.vax1 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.vax2 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.vax3 = (int *)calloc(indiv.Nindiv, sizeof(int));
    indiv.attend = (int *)calloc(indiv.Nindiv, sizeof(int));
    
    ReadFiles2(address_ind, &indiv);
    
    infecTimes = (int **)calloc(Nsim, sizeof(int *));
    for(i=0; i<Nsim; i++){
        infecTimes[i] = (int *)calloc(indiv.Nindiv, sizeof(int));
    }


    // Computations: -------------------------------------------------------------------------------------------------------------//
    
    AssignInfecTimes(&infecTimes, indiv, distr, Nsim, max_interval);



    // Saving results: -----------------------------------------------------------------------------------------------------------//
    
    FILE *infecFile;
    infecFile = fopen(address_output, "w");

    for(i=0; i<Nsim; i++){
        if(i>0){
            fprintf(infecFile, "\n");
        }
        for(j=0; j<indiv.Nindiv; j++){
            if(j<indiv.Nindiv-1){
                fprintf(infecFile, "%d ", infecTimes[i][j]);
            }
            else{
                fprintf(infecFile, "%d", infecTimes[i][j]);
            }
        }
    }

    fclose(infecFile);



    // Free: ---------------------------------------------------------------------------------------------------------------------//
   
    free(indiv.IDperson); 
    free(indiv.IDfamily); 
    free(indiv.IDcluster);
    free(indiv.symonset);
    free(indiv.diagnose);
    free(indiv.neg1);
    free(indiv.neg2);
    free(indiv.neg3);
    free(indiv.neg4);
    free(indiv.neg5);
    free(indiv.neg6);
    free(indiv.neg7);
    free(indiv.neg8); 
    free(indiv.neg9);
    free(indiv.vax1);
    free(indiv.vax2);
    free(indiv.vax3);
    free(indiv.attend); 

    for(i=0; i<Nsim; i++){
        free(infecTimes[i]);
    }
    free(infecTimes);


    return 0;
}



//--------------------------------------------------------------------------------------------------------------------------------//
// FUNCTIONS  --------------------------------------------------------------------------------------------------------------------//
//--------------------------------------------------------------------------------------------------------------------------------//

// Function: Read files
void ReadFiles1(char *address_glo, Indiv *indiv, Distr *distr){

    FILE *globalFile;
    globalFile = fopen(address_glo, "r");
    if(globalFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_glo); 
        exit(2);
    }
    while(!feof(globalFile)){
        fscanf(globalFile, "%d %lf %lf %lf %lf",
               &indiv->Nindiv,
               &distr->DISTR_INCUBATION_SHAPE,
               &distr->DISTR_INCUBATION_SCALE,
               &distr->DISTR_DIAGNOSIS_SHAPE,
               &distr->DISTR_DIAGNOSIS_SCALE);
    }
    fclose(globalFile);

}

// Function: Read files
void ReadFiles2(char *address_ind, Indiv *indiv){
    
    int i; i = 0;
    
    FILE *indivFile;
    indivFile = fopen(address_ind, "r");
    if(indivFile == NULL){
        fprintf(stderr, "\nERROR opening the file %s for reading\n", address_ind);
        exit(2);
    }
    while(!feof(indivFile)){
        fscanf(indivFile,"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d", 
               &indiv->IDperson[i],
               &indiv->IDfamily[i],
               &indiv->IDcluster[i],
               &indiv->symonset[i],
               &indiv->diagnose[i],
               &indiv->neg1[i],
               &indiv->neg2[i], 
               &indiv->neg3[i],
               &indiv->neg4[i],
               &indiv->neg5[i],
               &indiv->neg6[i],
               &indiv->neg7[i],
               &indiv->neg8[i], 
               &indiv->neg9[i],
               &indiv->vax1[i],
               &indiv->vax2[i],
               &indiv->vax3[i],
               &indiv->attend[i]); 
               i++;
    }
    fclose(indivFile);
}

// Function: Assign infection times
void AssignInfecTimes(int ***infecTimes, Indiv indiv, Distr distr, int Nsim, int max_interval){

    int i, j, k, l;

    // Pre-computation of all probabilities involved to sample infection times
    
    double *probPosTest, *probIncubation, *probDiagnosis, beta1, beta2, beta3, C, x, y, steps;
    probPosTest = (double *)calloc(max_interval, sizeof(double));
    probIncubation = (double *)calloc(max_interval, sizeof(double));
    probDiagnosis = (double *)calloc(max_interval, sizeof(double));
    beta1 = 1.51; beta2 = 2.19; beta3 = -1.1; C = 3.18;
    steps = 500.;
    
    for(i=0; i<max_interval; i++){
        x = i-C; // i = T_{P,n} - T_I
        if(x<0){y = beta1+beta2*x;} else{y = beta1+beta2*x+beta2*beta3*x;}
        probPosTest[i] = exp(y)/(exp(y)+1);
        for(j=0; j<steps; j++){
            probIncubation[i] += gsl_ran_gamma_pdf((double)(i+j/steps), distr.DISTR_INCUBATION_SHAPE, distr.DISTR_INCUBATION_SCALE)/steps;
            probDiagnosis[i] += gsl_ran_gamma_pdf((double)(i+j/steps), distr.DISTR_DIAGNOSIS_SHAPE, distr.DISTR_DIAGNOSIS_SCALE)/steps;
        }
    }
    

    // Computation of infection times

    int *negTests, timeInfec, NnegTests, diff, select, count;
    double *probTimeInfec, prod;
    gsl_ran_discrete_t *prob;
    
    prob = (gsl_ran_discrete_t *)calloc(max_interval, sizeof(gsl_ran_discrete_t));

    for(i=0; i<Nsim; i++){
        
        for(j=0; j<indiv.Nindiv; j++){

            NnegTests = 0;
            negTests = (int *)calloc(9, sizeof(int)); 
            probTimeInfec = (double *)calloc(max_interval, sizeof(double));

            if(indiv.neg1[j] != -1 && indiv.neg1[j] < indiv.diagnose[j]){NnegTests++; negTests[0] = indiv.neg1[j];} 
            if(indiv.neg2[j] != -1 && indiv.neg2[j] < indiv.diagnose[j]){NnegTests++; negTests[1] = indiv.neg2[j];} 
            if(indiv.neg3[j] != -1 && indiv.neg3[j] < indiv.diagnose[j]){NnegTests++; negTests[2] = indiv.neg3[j];} 
            if(indiv.neg4[j] != -1 && indiv.neg4[j] < indiv.diagnose[j]){NnegTests++; negTests[3] = indiv.neg4[j];} 
            if(indiv.neg5[j] != -1 && indiv.neg5[j] < indiv.diagnose[j]){NnegTests++; negTests[4] = indiv.neg5[j];} 
            if(indiv.neg6[j] != -1 && indiv.neg6[j] < indiv.diagnose[j]){NnegTests++; negTests[5] = indiv.neg6[j];}
            if(indiv.neg7[j] != -1 && indiv.neg7[j] < indiv.diagnose[j]){NnegTests++; negTests[6] = indiv.neg7[j];} 
            if(indiv.neg8[j] != -1 && indiv.neg8[j] < indiv.diagnose[j]){NnegTests++; negTests[7] = indiv.neg8[j];} 
            if(indiv.neg9[j] != -1 && indiv.neg9[j] < indiv.diagnose[j]){NnegTests++; negTests[8] = indiv.neg9[j];}

            if(indiv.symonset[j] != -1){
                for(k=0; k<max_interval; k++){
                    timeInfec = indiv.diagnose[j] - k;
                    prod = 1;
                    if(timeInfec <= indiv.symonset[j]){
                        for(l=0; l<NnegTests; l++){
                            diff = negTests[l]-timeInfec;
                            if(diff >= 0 && diff < max_interval){prod = prod * (1 - probPosTest[(int)(diff)]);}
                        }
                        if((int)(indiv.symonset[j] - timeInfec) >= max_interval){probTimeInfec[k] = 0;} else{
                        probTimeInfec[k] = probPosTest[k] * prod * probIncubation[(int)(indiv.symonset[j] - timeInfec)];}
                    }
                }
            }

            if(indiv.symonset[j] == -1){
                for(k=0; k<max_interval; k++){
                    timeInfec = indiv.diagnose[j] - k;
                    prod = 1;
                    for(l=0; l<NnegTests; l++){
                        diff = negTests[l]-timeInfec;
                        if(diff >= 0 && diff < max_interval){prod = prod * (1 - probPosTest[(int)(diff)]);}
                    }
                    probTimeInfec[k] = probPosTest[k] * prod * probDiagnosis[k];
                }
            }



            count = 0;
            for(k=0; k<max_interval; k++){
                if(probTimeInfec[k] > 0){count = 1;}
            }

            if(count == 1){
                prob = gsl_ran_discrete_preproc(max_interval, probTimeInfec); // trasforms probTimeInfec in a multinomial probability vector.
                select = gsl_ran_discrete(RNG, prob);  //printf("%d\n", select);
                gsl_ran_discrete_free(prob);

                free(negTests);
                free(probTimeInfec);
            
                (*infecTimes)[i][j] = (int)(indiv.diagnose[j] - select);
            } else{
                printf("\rInfection time not available for individual index %d. To be analysed.\n",j);
                (*infecTimes)[i][j] = (int)(indiv.diagnose[j]);
            }
            
        }
    }

    free(probPosTest); free(probIncubation); free(probDiagnosis);

}


